/*
 * Copyright (C) 2002 Robert Ernst <robert.ernst@linux-solutions.at>
 *
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See COPYING for GPL licensing information.
 *
 */

#include <qpe/qpeapplication.h>
#include <qtimer.h>
#include <qpainter.h>
#include <qwidget.h>
#include <qpixmap.h>
#include <qimage.h>
#include "Board.h"
#include <stdlib.h>
#include <unistd.h>

#include "layouts/data.c"

#define MIN(a, b) ((a < b) ? a : b)
#define MAX(a, b) ((a > b) ? a : b)
#define MOVE(layer, row, col) ((((m_tile[layer][row][col] & ~marked) & 0xFF) << 24) \
			      | ((layer & 0xFF) << 16) | ((row & 0xFF) << 8) | (col & 0xFF))
#define TILE(move) m_tile[(move & 0x00FF0000) >> 16][(move & 0xFF00) >> 8][move & 0xFF]
#define WHICH(move) ((move & 0xFF000000) >> 24)
#define POS(move) (move & 0x00FFFFFF)

Board::Board(QWidget *parent, const char *name, WFlags f) :
    QWidget(parent, name, f)
{
    m_sprites = new SpriteCollection(this);
    m_hinting = false;
    m_cheated = false;
    m_layout = 0;
    m_seconds = 0;
    memset(m_tile, 0, sizeof (m_tile));
    m_timer = new QTimer(this);
    connect(m_timer, SIGNAL(timeout()), this, SLOT(timerTick()));
}

Board::~Board()
{
    if (m_timer) {
	delete m_timer;
    }
    if (m_sprites) {
	delete m_sprites;
    }
}

void Board::displayHint(void)
{
    int move1;
    int move2;

    if (findHint(move1, move2)) {
	if (m_moves % 2) {
	    TILE(move2) = TILE(move2) | marked;
	} else {
	    TILE(move1) = TILE(move1) | marked;
	    TILE(move2) = TILE(move2) | marked;
	}
	repaint(false);
	m_hinting = true;
	int i;
	for (i = 0; i < 40; i++) {
	    qApp->processEvents();
	    usleep(25000);
	    if (!m_hinting) {
		break;
	    }
	}
	m_hinting = false;
	if (m_moves % 2) {
	    TILE(move2) = TILE(move2) & ~marked;
	} else {
	    TILE(move1) = TILE(move1) & ~marked;
	    TILE(move2) = TILE(move2) & ~marked;
	}
	repaint(false);
    } else {
	emit gameOver();
    }
}

void Board::autoPlay(void)
{
    int move1;
    int move2;

    if (m_moves % 2) {
	return;
    }
    while (1) {
	if (findHint(move1, move2)) {
	    TILE(move1) = TILE(move1) | marked;
	    TILE(move2) = TILE(move2) | marked;
	    repaint(false);
	    m_hinting = true;
	    int i;
	    for (i = 0; i < 40; i++) {
		qApp->processEvents();
		usleep(2500);
		if (!m_hinting) {
		    break;
		}
	    }
	    if (m_hinting) {
		m_hinting = false;
		TILE(move1) = 0;
		TILE(move2) = 0;
		m_move[m_moves] = move1;
		m_moves++;
		m_move[m_moves] = move2;
		m_moves++;
		repaint(false);
		if (m_moves >= moves) {
		    emit gameOver();
		    break;
		}
	    } else {
		TILE(move1) = TILE(move1) & ~marked;
		TILE(move2) = TILE(move2) & ~marked;
		repaint(false);
		break;
	    }
	} else {
	    emit gameOver();
	    break;
	}
    }
}

void Board::setCheated(bool cheated)
{
    m_cheated = cheated;
    repaint(0, 0, width(), 22, false);
}

bool Board::canUndo(void)
{
    return (!m_hinting && m_moves >= 2) ? true : false;
}

bool Board::canRedo(void)
{
    return (!m_hinting && ((m_maxmoves & ~0x01) - m_moves) >= 2) ? true : false;
}

void Board::undo(void)
{
    if (m_moves % 2) {
	TILE(m_move[m_moves - 1]) = TILE(m_move[m_moves - 1]) & ~marked;
	m_moves--;
	m_maxmoves = m_moves;
    }
    if (m_moves >= 2) {
	TILE(m_move[m_moves - 1]) = WHICH(m_move[m_moves - 1]);
	TILE(m_move[m_moves - 2]) = WHICH(m_move[m_moves - 2]);
	m_moves = m_moves - 2;
    }
    repaint(false);
    emit changed();
}

void Board::redo(void)
{
    if ((m_moves % 2) == 0 && (m_maxmoves - m_moves) >= 2) {
	TILE(m_move[m_moves]) = 0;
	TILE(m_move[m_moves + 1]) = 0;
	m_moves = m_moves + 2;
	repaint(false);
	emit changed();
    }
}

bool Board::tilesLeft(void)
{
    return (m_moves < moves) ? true : false;
}

void Board::updateLayout(int layout)
{
    m_layout = layout;
}

void Board::start(void)
{
    makeBoard(m_layout);
    m_hinting = false;
    m_cheated = false;
    m_moves = 0;
    m_maxmoves = 0;
    m_seconds = 0;
    memset(m_move, 0, sizeof (m_move));
    repaint(false);
}

void Board::timerTick(void)
{
    m_seconds++;
    repaint(false);
}

void Board::showEvent(QShowEvent *event)
{
    QWidget::showEvent(event);
    m_timer->start(1000);
    setFocusPolicy(StrongFocus);
}

void Board::hideEvent(QHideEvent *event)
{
    QWidget::hideEvent(event);
    m_timer->stop();
}

void Board::paintEvent(QPaintEvent *event)
{
    QPixmap pixmap(width(), height());
    QPainter painter_widget(this);
    QPainter painter_pixmap(&pixmap);
    int i;
    int j;
    int k;

    painter_pixmap.drawPixmap(0, 0, *(m_sprites->backgroundPixmap()));
    for (k = 0; k < layers; k++) {
	for (j = 0; j < rows; j++) {
	    for (i = 0; i < cols; i++) {
		if (!m_tile[k][j][i]) {
		    continue;
		}
		int xpos = xBorder() + i * (m_sprites->tileWidth() / 2) + k * 2;
		int ypos = yBorder() + 22 + j * (m_sprites->tileHeight() / 2) - k * 2;
		QRect rect(xpos, ypos, m_sprites->tileWidth(), m_sprites->tileHeight());
		if (!event->rect().intersects(rect)) {
		    continue;
		}
		if (m_tile[k][j][i] & marked) {
		    int which = (m_tile[k][j][i] & ~marked) - 1;
		    QPixmap *lpm = lighten(m_sprites->tilePixmap(which));
		    painter_pixmap.drawPixmap(xpos, ypos, *lpm);
		    delete lpm;
		} else {
		    int which = m_tile[k][j][i] - 1;
		    painter_pixmap.drawPixmap(xpos, ypos, *m_sprites->tilePixmap(which));
		}
	    }
	}
    }

    int seconds = getTimeForGame();
    QString s;
    s.sprintf("%02d:%02d:%02d%s", seconds / 3600, (seconds / 60) % 60, seconds % 60,
	      m_cheated ? + tr(" (cheating)") : "");
    painter_pixmap.fillRect(0, 0, width(), 22, backgroundColor());
    painter_pixmap.setPen(QColor(0, 0, 0));
    painter_pixmap.drawText(7, 0, width(), 22, AlignLeft | AlignVCenter, s);

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, pixmap);
}

void Board::mousePressEvent(QMouseEvent *event)
{
    int layer;
    int row;
    int col;

    if (m_hinting) {
	m_hinting = false;
	return;
    }
    if (findTile(event->x(), event->y(), layer, row, col)
	&& tileIsFree(layer, row, col)) {
	if (m_tile[layer][row][col] & marked) {
	    m_tile[layer][row][col] = m_tile[layer][row][col] & ~marked;
	    m_moves--;
	    m_maxmoves = m_moves;
	    repaint(false);
	    emit changed();
	    return;
	}
	m_tile[layer][row][col] = m_tile[layer][row][col] | marked;
	m_move[m_moves] = MOVE(layer, row, col);
	m_moves++;
	m_maxmoves = m_moves;
	repaint(false);
	emit changed();
	if ((m_moves % 2) == 0) {
	    if (TILE(m_move[m_moves - 1]) == TILE(m_move[m_moves - 2])) {
		int i;
		m_hinting = true;
		for (i = 0; i < 10; i++) {
		    qApp->processEvents();
		    usleep(25000);
		    if (!m_hinting) {
			break;
		    }
		}
		m_hinting = false;
		TILE(m_move[m_moves - 1]) = 0;
		TILE(m_move[m_moves - 2]) = 0;
		repaint(false);
		if (m_moves >= moves || noMovePossible()) {
		    emit gameOver();
		} else {
		    emit changed();
		}
	    } else {
		TILE(m_move[m_moves - 1]) = TILE(m_move[m_moves - 1]) & ~marked;
		TILE(m_move[m_moves - 2]) = TILE(m_move[m_moves - 2]) & ~marked;
		m_moves = m_moves - 2;
		m_maxmoves = m_moves;
		repaint(false);
		emit changed();
	    }
	}
    }
}

int Board::getTimeForGame(void)
{
    return m_seconds;
}

int Board::xBorder(void)
{
    return (width() - ((cols + 1) / 2) * m_sprites->tileWidth()) / 2;
}

int Board::yBorder(void)
{
    return (height() - 22 - ((rows + 1) / 2) * m_sprites->tileHeight()) / 2;
}

bool Board::findTile(int x, int y, int &layer, int &row, int &col)
{
    if (x < xBorder()
	|| x >= (xBorder() + (cols + 1) * (m_sprites->tileWidth() / 2) + (layers - 1) * 2)
	|| y < (yBorder() + 22 - (layers - 1) * 2)
	|| y >= (yBorder() + 22 + (rows + 1) * (m_sprites->tileHeight() / 2))) {
	return false;
    }
    int tmp_layer;
    for (tmp_layer = layers - 1; tmp_layer >= 0; tmp_layer--) {
	int tmp_col = (x - xBorder() - 2 * tmp_layer) / (m_sprites->tileWidth() / 2);
	int tmp_row = (y - yBorder() + 2 * tmp_layer - 22) / (m_sprites->tileHeight() / 2);
	if (tmp_col >= cols) {
	    tmp_col = cols - 1;
	} else if (tmp_col < 0) {
	    tmp_col = 0;
	}
	if (tmp_row >= rows) {
	    tmp_row = rows - 1;
	} else if (tmp_row < 0) {
	    tmp_row = 0;
	}
	if (m_tile[tmp_layer][tmp_row][tmp_col]) {
	    col = tmp_col;
	    row = tmp_row;
	    layer = tmp_layer;
	    return true;
	} else if (tmp_row > 0 && tmp_col > 0 && m_tile[tmp_layer][tmp_row - 1][tmp_col - 1]) {
	    col = tmp_col - 1;
	    row = tmp_row - 1;
	    layer = tmp_layer;
	    return true;
	} else if (tmp_row > 0 && m_tile[tmp_layer][tmp_row - 1][tmp_col]) {
	    col = tmp_col;
	    row = tmp_row - 1;
	    layer = tmp_layer;
	    return true;
	} else if (tmp_col > 0 && m_tile[tmp_layer][tmp_row][tmp_col - 1]) {
	    col = tmp_col - 1;
	    row = tmp_row;
	    layer = tmp_layer;
	    return true;
	}
    }
    return false;
}

bool Board::tileIsFree(int layer, int row, int col)
{
    if (col > 1 && col < (cols - 2)) {
	bool left_is_free = true;
	if (row > 0 && m_tile[layer][row - 1][col - 2]) {
	    left_is_free = false;
	} else if (m_tile[layer][row][col - 2]) {
	    left_is_free = false;
	} else if (row < (rows - 1) && m_tile[layer][row + 1][col - 2]) {
	    left_is_free = false;
	}
	bool right_is_free = true;
	if (row > 0 && m_tile[layer][row - 1][col + 2]) {
	    right_is_free = false;
	} else if (m_tile[layer][row][col + 2]) {
	    right_is_free = false;
	} else if (row < (rows - 1) && m_tile[layer][row + 1][col + 2]) {
	    right_is_free = false;
	}
	if (!left_is_free && !right_is_free) {
	    return false;
	}
    }
    while (layer < (layers - 1)) {
	layer++;
	int tmp_row;
	int tmp_col;
	for (tmp_row = row - 1; tmp_row <= (row + 1); tmp_row++) {
	    for (tmp_col = col - 1; tmp_col <= (col + 1); tmp_col++) {
		if (tmp_row >= 0 && tmp_col >= 0 && tmp_row < rows && tmp_col < cols
		    && m_tile[layer][tmp_row][tmp_col]) {
		    return false;
		}
	    }
	}
    }
    return true;
}

QPixmap *Board::lighten(QPixmap *src)
{
    const float FACTOR = 1.3;

    QImage img = src->convertToImage();
    if(img.depth() > 8) { // at least high-color
	for(int y = 0; y < src->height(); y++) {
	    uchar *p = (uchar *)img.scanLine(y);
	    for(int x = 0; x < src->width() * 4; x++) {
		*p = (unsigned char)MIN(255, (int)(FACTOR * (*p)));
		p++;
	    }
	}
    } else { // image has a palette
	// get background color index
	int idx = img.pixelIndex(8, 1); // should work for all tiles
	img.setColor(idx, QColor(img.color(idx)).light().rgb());
    }
 
    QPixmap *pm = new QPixmap();
    pm->convertFromImage(img);
    return pm;
}

void Board::makeBoard(int layout)
{
    bool used[moves];
    bool set[layers][rows][cols];
    int tiles_done;
    int layer;
    int row;
    int col;
    int nr;
    const char **ptr;

    memset(m_tile, 0, sizeof (m_tile));
    memset(used, 0, sizeof (used));
    memset(set, 0, sizeof (set));
    switch (layout) {
    case 0: ptr = layout_1; break;
    case 1: ptr = layout_2; break;
    case 2: ptr = layout_3; break;
    case 3: ptr = layout_4; break;
    case 4: ptr = layout_5; break;
    default:
	ptr = layout_1;
    }
    for (layer = 0; layer < layers; layer++) {
	for (row = 0; row < rows; row++) {
	    int idx = 1 + layer * (rows + 1) + row;
	    const char *ptr_row = ptr[idx];
	    for (col = 0; col < cols; col++) {
		if (ptr_row[col] == '1') {
		    set[layer][row][col] = true;
		}
	    }
	}
    }
    for (tiles_done = 0; tiles_done < moves; tiles_done++) {
	do {
	    nr = rand() % moves;
	} while (used[nr]);
	used[nr] = true;
	do {
	    layer = rand() % layers;
	    row = rand() % rows;
	    col = rand() % cols;
	} while (!set[layer][row][col] || !canPlace(layer, row, col));
	set[layer][row][col] = false;
	m_tile[layer][row][col] = tileIndex(nr) + 1;
    }
}

bool Board::canPlace(int layer, int row, int col)
{
    if (layer > 0) {
	int quarters = 0;
	if (m_tile[layer - 1][row][col]) {
	    quarters = 4;
	} else {
	    if (col > 0) {
		if (m_tile[layer - 1][row][col - 1]) {
		    quarters = quarters + 2;
		} else {
		    if (row > 0 && m_tile[layer - 1][row - 1][col - 1]) {
			quarters++;
		    }
		    if (row < (rows - 1) && m_tile[layer - 1][row + 1][col - 1]) {
			quarters++;
		    }
		}
	    }
	    if (col < (cols - 1)) {
		if (m_tile[layer - 1][row][col + 1]) {
		    quarters = quarters + 2;
		} else {
		    if (row > 0 && m_tile[layer - 1][row - 1][col + 1]) {
			quarters++;
		    }
		    if (row < (rows - 1) && m_tile[layer - 1][row + 1][col + 1]) {
			quarters++;
		    }
		}
	    }
	    if (row > 0 && m_tile[layer - 1][row - 1][col]) {
		quarters = quarters + 2;
	    }
	    if (row < (rows - 1) && m_tile[layer - 1][row + 1][col]) {
		quarters = quarters + 2;
	    }
	}
	if (quarters < 4) {
	    return false;
	}
    }
    return true;
}

bool Board::noMovePossible(void)
{
    int move1;
    int move2;

    return !findHint(move1, move2);
}

bool Board::findHint(int &move1, int &move2)
{
    if (m_moves % 2) {
	move1 = m_move[m_moves - 1];
	return findHintFor(move1, move2);
    } else {
	int layer;
	int row;
	int col;
	for (row = 0; row < rows; row++)  {
	    for (col = 0; col < cols; col++) {
		for (layer = layers - 1; layer >= 0; layer--) {
		    if (m_tile[layer][row][col] != 0
			&& tileIsFree(layer, row, col)) {
			move1 = MOVE(layer, row, col);
			if (findHintFor(move1, move2)) {
			    return true;
			}
		    }
		}
	    }
	}
	return false;
    }
}

bool Board::findHintFor(int move1, int &move2)
{
    int layer;
    int row;
    int col;

    for (layer = layers - 1; layer >= 0; layer--) {
	for (row = 0; row < rows; row++) {
	    for (col = 0; col < cols; col++) {
		move2 = MOVE(layer, row, col);
		if (POS(move1) != POS(move2)
		    && (m_tile[layer][row][col] & ~marked) == (int)WHICH(move1)
		    && tileIsFree(layer, row, col)) {
		    return true;
		}
	    }
	}
    }
    return false;
}

int Board::tileIndex(int nr)
{
    nr = nr / 4;
    if (nr < 35) {
	return nr;
    } else if (nr < 42) {
	return nr + 1;
    } else {
	return 44;
    }
}

