/*
 * Copyright (C) 2002 Robert Ernst <robert.ernst@linux-solutions.at>
 *
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See COPYING for GPL licensing information.
 *
 */

#include <qpe/qpemenubar.h>
#include <qpe/config.h>
#include <qapplication.h>
#include <qmainwindow.h>
#include <qpopupmenu.h>
#include <qmessagebox.h>
#include "Mahjongg.h"
#include "Board.h"
#include <stdlib.h>
#include <time.h>

Mahjongg::Mahjongg(QWidget *parent, const char *name, WFlags f) :
    QMainWindow(parent, name, f)
{
    /* initialize random number generator */
    srand(::time(0));

    /* set windows style and caption */
    setCaption("Mahjongg");

    /* create the play field as the central widget */
    m_board = new Board(this);
    setCentralWidget(m_board);
    m_board->show();
    connect(m_board, SIGNAL(changed()), this, SLOT(enableItems()));
    connect(m_board, SIGNAL(gameOver()), this, SLOT(gameOver()));

    /* create the game/layout submenu */
    m_layoutMenu = new QPopupMenu(0, "layout_menu");
    m_layoutMenu->setCheckable(true);
    connect(m_layoutMenu, SIGNAL(activated(int)), this, SLOT(updateLayout(int)));
    m_layoutMenu->insertItem(tr("Mahjongg"), 0);
    m_layoutMenu->insertItem(tr("Pyramid"), 1);
    m_layoutMenu->insertItem(tr("Towers"), 2);
    m_layoutMenu->insertItem(tr("Triangle"), 3);
    m_layoutMenu->insertItem(tr("Walls"), 4);

    /* create the game submenu */
    m_gameMenu = new QPopupMenu(0, "game_menu");
    m_gameMenu->insertItem(tr("New"), this, SLOT(start()));
    m_gameMenu->insertItem(tr("AutoPlay"), m_board, SLOT(autoPlay()));
    m_gameMenu->insertItem(tr("Layout"), m_layoutMenu);
    m_gameMenu->insertSeparator();
    m_gameMenu->insertItem(tr("Hint"), this, SLOT(hint()));
    m_gameMenu->insertItem(tr("Undo"), this, SLOT(undo()), 0, 0x0100);
    m_gameMenu->insertItem(tr("Redo"), this, SLOT(redo()), 0, 0x0101);
    m_gameMenu->insertSeparator();
    m_gameMenu->insertItem(tr("Close"), this, SLOT(close()));

    /* create the menu */
    QPEMenuBar *menu = new QPEMenuBar(this);
    menu->insertItem(tr("Game"), m_gameMenu);

    /* read the configuration file */
    readConfig();

    /* check the menu entries of game size and game iq submenu */
    m_layoutMenu->setItemChecked(m_layout, true);
    m_board->updateLayout(m_layout);
    enableItems();

    /* start a game automatically */
    start();
}

Mahjongg::~Mahjongg()
{
    writeConfig();
}

void Mahjongg::readConfig(void)
{
    Config cfg("Mahjongg");

    cfg.setGroup("Settings");

    int layout = cfg.readNumEntry("Layout", 0);
    if (layout <= 0) {
	m_layout = 0;
    } else if (layout >= 4) {
	m_layout = 4;
    } else {
	m_layout = layout;
    }
    m_board->updateLayout(m_layout);
}

void Mahjongg::writeConfig(void)
{
    Config cfg("Mahjongg");

    cfg.setGroup("Settings");
    cfg.writeEntry("Layout", m_layout);
}

void Mahjongg::updateLayout(int layout)
{
    m_layoutMenu->setItemChecked(m_layout, false);
    m_layout = layout;
    m_layoutMenu->setItemChecked(m_layout, true);
    writeConfig();
    m_board->updateLayout(layout);
    start();
}

void Mahjongg::start(void)
{
    m_board->start();
    enableItems();
}

void Mahjongg::hint(void)
{
    qApp->processEvents();
    m_board->displayHint();
    m_board->setCheated(true);
    enableItems();
}

void Mahjongg::undo(void)
{
    if (m_board->canUndo()) {
	m_board->undo();
	m_board->setCheated(true);
	enableItems();
    }
}

void Mahjongg::redo(void)
{
    if(m_board->canRedo()) {
	m_board->redo();
	enableItems();
    }
}

void Mahjongg::enableItems(void)
{
    m_gameMenu->setItemEnabled(0x100, m_board->canUndo());
    m_gameMenu->setItemEnabled(0x101, m_board->canRedo());
}

void Mahjongg::gameOver(void)
{
    if (m_board->tilesLeft()) {
	QMessageBox::information(this, tr("Mahjongg"),
				 tr("No more moves\npossible\nYou loose"));
    } else {
	int seconds = m_board->getTimeForGame();
	QString s;
	s.sprintf("%02d:%02d:%02d", seconds / 3600, (seconds / 60) % 60, seconds % 60);
	QMessageBox::information(this, tr("Mahjongg"),
				 tr("Congratulations!\nYou made it in\n%1").arg(s));
    }
    start();
}

